@@ -451,18 +451,6 @@ struct PutItem<'a> {
     return_values_on_condition_check_failure: Option<ReturnValues>,
 }
 
-/// A DynamoDB [GetItem] payload
-///
-/// [GetItem]: https://docs.aws.amazon.com/amazondynamodb/latest/APIReference/API_GetItem.html
-#[derive(Serialize)]
-#[serde(rename_all = "PascalCase")]
-struct GetItem<'a> {
-    /// The table name
-    table_name: &'a str,
-    /// The primary key
-    key: Map<'a, &'a str, AttributeValue<'a>>,
-}
-
 #[derive(Deserialize)]
 struct ErrorResponse<'a> {
     #[serde(rename = "__type")]
