@@ -16,7 +16,7 @@
 // under the License.
 
 use std::future::Future;
-use std::time::Instant;
+use std::time::{Duration, Instant};
 use tokio::sync::Mutex;
 
 /// A temporary authentication token with an associated expiry
@@ -34,17 +34,25 @@ pub struct TemporaryToken<T> {
 #[derive(Debug)]
 pub struct TokenCache<T> {
     cache: Mutex<Option<TemporaryToken<T>>>,
+    min_ttl: Duration,
 }
 
 impl<T> Default for TokenCache<T> {
     fn default() -> Self {
         Self {
             cache: Default::default(),
+            min_ttl: Duration::from_secs(300),
         }
     }
 }
 
 impl<T: Clone + Send> TokenCache<T> {
+    /// Override the minimum remaining TTL for a cached token to be used
+    #[cfg(feature = "aws")]
+    pub fn with_min_ttl(self, min_ttl: Duration) -> Self {
+        Self { min_ttl, ..self }
+    }
+
     pub async fn get_or_insert_with<F, Fut, E>(&self, f: F) -> Result<T, E>
     where
         F: FnOnce() -> Fut + Send,
@@ -55,13 +63,7 @@ impl<T: Clone + Send> TokenCache<T> {
 
         if let Some(cached) = locked.as_ref() {
             match cached.expiry {
-                Some(ttl)
-                    if ttl
-                        .checked_duration_since(now)
-                        .unwrap_or_default()
-                        .as_secs()
-                        > 300 =>
-                {
+                Some(ttl) if ttl.checked_duration_since(now).unwrap_or_default() > self.min_ttl => {
                     return Ok(cached.token.clone());
                 }
                 None => return Ok(cached.token.clone()),
