@@ -19,12 +19,11 @@
 use bytes::Bytes;
 use futures::{stream::BoxStream, StreamExt, TryStreamExt};
 use std::ops::Range;
-use tokio::io::AsyncWrite;
 
 use crate::path::Path;
 use crate::{
-    GetOptions, GetResult, ListResult, MultipartId, ObjectMeta, ObjectStore, PutOptions, PutResult,
-    Result,
+    GetOptions, GetResult, ListResult, MultipartUpload, ObjectMeta, ObjectStore, PutMultipartOpts,
+    PutOptions, PutPayload, PutResult, Result,
 };
 
 #[doc(hidden)]
@@ -81,28 +80,35 @@ impl<T: ObjectStore> PrefixStore<T> {
 
 #[async_trait::async_trait]
 impl<T: ObjectStore> ObjectStore for PrefixStore<T> {
-    async fn put(&self, location: &Path, bytes: Bytes) -> Result<PutResult> {
+    async fn put(&self, location: &Path, payload: PutPayload) -> Result<PutResult> {
         let full_path = self.full_path(location);
-        self.inner.put(&full_path, bytes).await
+        self.inner.put(&full_path, payload).await
     }
 
-    async fn put_opts(&self, location: &Path, bytes: Bytes, opts: PutOptions) -> Result<PutResult> {
-        let full_path = self.full_path(location);
-        self.inner.put_opts(&full_path, bytes, opts).await
-    }
-
-    async fn put_multipart(
+    async fn put_opts(
         &self,
         location: &Path,
-    ) -> Result<(MultipartId, Box<dyn AsyncWrite + Unpin + Send>)> {
+        payload: PutPayload,
+        opts: PutOptions,
+    ) -> Result<PutResult> {
+        let full_path = self.full_path(location);
+        self.inner.put_opts(&full_path, payload, opts).await
+    }
+
+    async fn put_multipart(&self, location: &Path) -> Result<Box<dyn MultipartUpload>> {
         let full_path = self.full_path(location);
         self.inner.put_multipart(&full_path).await
     }
 
-    async fn abort_multipart(&self, location: &Path, multipart_id: &MultipartId) -> Result<()> {
+    async fn put_multipart_opts(
+        &self,
+        location: &Path,
+        opts: PutMultipartOpts,
+    ) -> Result<Box<dyn MultipartUpload>> {
         let full_path = self.full_path(location);
-        self.inner.abort_multipart(&full_path, multipart_id).await
+        self.inner.put_multipart_opts(&full_path, opts).await
     }
+
     async fn get(&self, location: &Path) -> Result<GetResult> {
         let full_path = self.full_path(location);
         self.inner.get(&full_path).await
@@ -198,9 +204,8 @@ impl<T: ObjectStore> ObjectStore for PrefixStore<T> {
 #[cfg(test)]
 mod tests {
     use super::*;
+    use crate::integration::*;
     use crate::local::LocalFileSystem;
-    use crate::test_util::flatten_list_stream;
-    use crate::tests::*;
 
     use tempfile::TempDir;
 
@@ -226,9 +231,8 @@ mod tests {
 
         let location = Path::from("prefix/test_file.json");
         let data = Bytes::from("arbitrary data");
-        let expected_data = data.clone();
 
-        local.put(&location, data).await.unwrap();
+        local.put(&location, data.clone().into()).await.unwrap();
 
         let prefix = PrefixStore::new(local, "prefix");
         let location_prefix = Path::from("test_file.json");
@@ -247,11 +251,11 @@ mod tests {
             .bytes()
             .await
             .unwrap();
-        assert_eq!(&*read_data, expected_data);
+        assert_eq!(&*read_data, data);
 
         let target_prefix = Path::from("/test_written.json");
         prefix
-            .put(&target_prefix, expected_data.clone())
+            .put(&target_prefix, data.clone().into())
             .await
             .unwrap();
 
@@ -264,6 +268,6 @@ mod tests {
 
         let location = Path::from("prefix/test_written.json");
         let read_data = local.get(&location).await.unwrap().bytes().await.unwrap();
-        assert_eq!(&*read_data, expected_data)
+        assert_eq!(&*read_data, data)
     }
 }
